
/*This a FEM program for constant strain triangular elements*/



#include <stdio.h>

#include <stdlib.h>

#include <string.h>

#include <math.h>

#include <time.h>

#define NE 200

#define MM 25

#define BUFF 256

FILE *fp2,*fp3;

char buf[BUFF];

int nn,ne,nb,nm,nl,ndime,nstr,nbe,kmax,nmax,iln,nq;

int *nod, nco[MM][4], mat[5],nc[5];

int nsb;

double dt,damp;

double *cord;

double *ste;

double amat[5][10],aload[5][3];

double est[6][6],s[NE][3][6];

int nn2;

double *ok, *f;



void fst(int n);


main()
{
time_t t1, t2;
char file_sav[30];
/*-------------------- Read in the dat file ----------------------*/

rinput(file_sav);


 /* Opening the data output file  for FEM*/

time(&t1);

openfile_fp2(file_sav);

ste = calloc(ne*3, sizeof(double));
if(ste == NULL)
{printf(" out of core for ste\n");exit(1);}



/*----------------------- F. E. Program --------------------------- */

  openfile_fp3(file_sav); /* open for write the stress output file*/

  semi_band();       /* calc nsb */

  printf("\nSemi-Band-Width = %d\n",nsb);

  nn2 = 2*nn;

  ok = calloc(nn2*nsb, sizeof(double));
  if(ok == NULL)
  {printf(" out of core for ok\n");exit(1);}

  f = calloc(nn2, sizeof(double));
  if(f == NULL)
  {printf(" out of core for f\n");exit(1);}



  elem();

time(&t2);
fprintf(fp2,"\n\n*** Time for analysis = %f min or %d sec ***\n",
       (float)(t2-t1)/60., t2-t1);
fprintf(stdout,"\n\n*** Time for analysis = %f min or %d sec ***\n",
       (float)(t2-t1)/60., t2-t1);
printf("*** The stresses are stored in %s.str ***\n",file_sav);
printf("*** The results of FE  are stored in %s.fem ***\n", file_sav);
fclose(fp3);

fclose(fp2);
fclose(fp3);

return;

}




/* Function openfile */
openfile_fp2(char *file_sav)
{
char file[30];
strcpy(file,file_sav); strcat(file,".fem");
fp2=fopen(file,"w");
	if(fp2==(FILE *)NULL)
	{
	printf("cannot open %s\n",file);
	exit(1);
	}
return;
}

openfile_fp3(char *file_sav)
{
char file[30];
strcpy(file, file_sav); strcat(file,".str");
fp3=fopen(file,"w");
	if(fp3==(FILE *)NULL)
	{
	printf("cannot open %s\n", file);
	exit(1);
	}
return;
}

/*** Finite Element function  ***/


elem()
{


printf("\n F E analysis started with node = %d and elements = %d\n",nn,ne);

   /* zero s, ok and f */
   zero_ok();

   /* put aload[] into f[] */
   aload_f();


  /* form stiff mat ok */
  printf("\n \n Assembling stiffness matrix \n");
  form_ok();


  /* apply boundary conds to the ok[][] */
  bc_ok();

  /* solution by gaussian elimination */
  printf("\n \n Solving stiffness matrix \n \n \n");
  gauss_el();

  /* print the nodal displacements */
  fpr_defl();

  /* calculate and fprint the stresses */
  calc_stress();

return;
}

/* Function to zero back subs matrix s, overall stiffness matrix ok
 and load vector f */
zero_ok()
{
int i,j,k,l;
	for(i=0; i< ne; i++)
	{
		for(k=0; k< 3; k++)
		{
			for(l=0; l< 6; l++) s[i][k][l]=0.;
		}
	}

	for(i=0; i< 2*nn; i++)
	{
		for(j=0; j< nsb; j++) ok[nsb*i + j]=0.;
	f[i]=0.;
	}
return;
}


/* function to read in applied aload()  into f(array), must have load
card for last node even if no applied load */
aload_f()
{
int i,ii,j,jj,k;

	if(iln >0 )
	{
		for(i=0; nc[i] < nn; i++)
		{
		ii=2*nc[i]-2;
		jj=2*nc[i]-1;
			for(k=0 ; k < 2; k++)
			{
			if(k==0) f[ii]=aload[i][k];
			else
			f[jj]=aload[i][k];
			}
		}
	}
return;
}


/* function to assemble the stiffness mat by adding the individual
   element matrices formed by function fst() */
form_ok()
{
int i,n,ni,j,nj,ja,jb,k,l, nrow, ncol;
/* loop for each element */

	for(n=0; n < ne; n++)
	{
	fst(n);
		for(i=0; i< 3; i++)
		{
		ni=2*(nod[4*n + i]-1);
			for(j=0; j< 3; j++)
			{
			nj=2*(nod[4*n + j]-1);
				for(k=0; k< 2; k++)
				{
					for(l=0; l< 2; l++)
					{
					ja = i*2 + k;
					jb = j*2 + l;
			  
					nrow = ni + k;
					ncol = nj + l;
					if(ncol < nrow) continue;
					ncol = ncol - nrow;

			  ok[nsb*nrow + ncol] += est[ja][jb];  

					}
				}
			}
		}
	}
			
return;
}


/* function applies boundry conditions to stiffness matrix ok and check 
each node in turn */
bc_ok()
{
int ip,il,nrow,nri;
	for(ip=0; ip< nb; ip++)
	{
	il=nco[ip][0];
	nrow=(il*2)-2;
	/* check in  x direction */
	if(nco[ip][1] < 1) goto lab1;
	ok[nsb*nrow + 0]=pow(10.,33);
	f[nrow]=0.;
lab1:   
	/* check in y direction  */
	if(nco[ip][2] < 1) continue;
	nri=nrow+1;
	ok[nsb*nri + 0]=pow(10.0,33);
	f[nri]=0.;
	}
return;
}


/* function for the solution of ok[][] by gaussian elimination */
gauss_el()
{
int m1,mm,i,ii,j,k,jk, nba1,l;
double fac,sum;

	mm = nn2;
	m1 = mm - 1;
	for(i = 0 ; i < m1 ; i++)
	{
	ii = i + 1;
	nba1 = nsb + i ;
	  for(j = ii; j < nba1; j++)
	  {

	  if(j > mm - 1) break;
	  fac=ok[nsb*i + (j - i)]/ok[nsb*i + 0];

	      for(k = 0;k < nsb; k++)
	      {
	      l = k + (j - i);
	      if(l > nsb - 1) continue;
	      ok[nsb*j + k] += -fac*ok[nsb*i + l];
	      }
	  f[j]=f[j]-fac*f[i];
	  }
	}

/* Back subsitution  */

	f[mm-1]=f[mm-1]/ok[nsb*(mm-1) + 0];
	for(i = 0 ; i < m1; i++)
	{
	j = mm - i - 2;
	sum = 0.;
		for(jk = 1; jk < nsb; jk++)
		{
		if((j + jk) > mm - 1)  goto skip_2;
		sum +=ok[nsb*j + jk]*f[j + jk];
		}
	skip_2:
	f[j] =(f[j] -sum)/ok[nsb*j + 0];
	}
return;
}

/* function to print the nodal displacements */
fpr_defl()
{
int i,l,m;

	m = 2*nn;

	fprintf(fp2,"\n \n        Results \
	   \n \n Deflections\n");
	fprintf(fp2,"   Node   x-def          y-def\n");
	
	for(i=0; i< m; i=i+2)
	{
	l=(i+2)/2;
	fprintf(fp2,"%5d   %15.8e %15.8e\n", l, f[i], f[i+1]);
	}
return;
}


/* function to calculate and fprint stresses */
calc_stress()
{
int n,l,i,j,k;
double def[6], cc,aa,smax,smin,ang;
	fprintf(fp2," \n \n Stresses \n");

	fprintf(fp2,"Elm  Sx           Sy           Sxy \
	  Smax         Smin         Angle\n");     

	for(n=0; n < ne; n++)
	{
		for(l=0 ;l < 3; l++)
		ste[3*n + l]=0. ;
		
		for(i=0; i < 3; i++)
		{
		def[i*2]=f[nod[4*n + i]*2-2];
		def[i*2+1]=f[nod[4*n + i]*2-1];
		}

		/* Multiply the stress matrix by def */

		for(j=0; j < 3; j++)
		{
			for(k=0; k< 6; k++)
			ste[3*n + j] += s[n][j][k]*def[k];
		}

		cc=(ste[3*n + 0]+ste[3*n + 1])/2. ;
		aa=sqrt(pow((ste[3*n + 0]-ste[3*n + 1]),2)/4.+
		pow(ste[3*n + 2],2));
		smax=cc+aa;
		smin=cc-aa;
		if(fabs(ste[3*n + 1] - smin) <= pow(10,-10))
		ang = 90.0;
		else
		ang=57.29578*atan(ste[3*n + 2]/(ste[3*n + 1]-smin));
		fprintf(fp2,"%3d %12.5e %12.5e %12.5e %12.5e\
 %12.5e %11.4e\n",n+1,ste[3*n + 0],ste[3*n + 1],ste[3*n + 2],smax,smin,ang);
		fprintf(fp3,"%12.5e%12.5e%12.5e%12.5e%12.5e\
\n",ste[3*n + 0],ste[3*n + 1],ste[3*n + 2],smax,smin);
	
	}
return;
}                
 



			/* FUNCTION FST */

void fst(int n)
{
int n1,n2,n3,ll,i,j,k;
double v,e,xn1,xn2,xn3,yn1,yn2,yn3,comm,b1,b2,b3,c1,c2,c3,area,thic;
double b[3][6], d[3][3];        
/* This function works out the element stiffnesses of triangles */

	/* get nodes out of nod */
	
	n1=nod[4*n + 0]-1;
	n2=nod[4*n + 1]-1;
	n3=nod[4*n + 2]-1;
	
	/* get material properties and thichnesses */
	
	ll=nod[4*n + 3]-1;
	thic=amat[ll][2];
	v=amat[ll][0];
	e=amat[ll][1];

	/* define coordinates */

	xn1=cord[2*n1 + 0];
	xn2=cord[2*n2 + 0];
	xn3=cord[2*n3 + 0];

	yn1=cord[2*n1 + 1];
	yn2=cord[2*n2 + 1];
	yn3=cord[2*n3 + 1];

	/* Form elasticity matrix D */
	form_D(v,e,d);

	/* form strain displacement matrix b */

	b1=yn2-yn3;
	b2=yn3-yn1;
	b3=yn1-yn2;

	c1=xn3-xn2;
	c2=xn1-xn3;
	c3=xn2-xn1;

	area=(-(c2*b2)+(2.*(c1*b3))+(c3*b3)+(c1*b1))/2.;

	b[0][0]=b1/(2.*area);
	b[0][1]=0.;
	b[0][2]=b2/(2.*area);
	b[0][3]=0.;
	b[0][4]=b3/(2.*area);
	b[0][5]=0.;

	b[1][0]=0.;
	b[1][1]=c1/(2.*area);
	b[1][2]=0.;
	b[1][3]=c2/(2.*area);
	b[1][4]=0.;
	b[1][5]=c3/(2.*area);

	b[2][0]=c1/(2.*area);
	b[2][1]=b1/(2.*area);
	b[2][2]=c2/(2.*area);
	b[2][3]=b2/(2.*area);
	b[2][4]=c3/(2.*area);
	b[2][5]=b3/(2.*area);

	/* form back subsitution matrix */

	for(j=0; j< 6; j++)
	{
		for(i=0; i< 3; i++)
		{
			for(k=0; k< 3; k++)
			s[n][i][j] += d[i][k]*b[k][j];
		}
	}
	
	/* form element stiffness */

	for(i=0; i < 6; i++)
	{
		for(j = 0; j < 6; j++)
		{
		est[i][j]=0.;
			for(k=0; k < 3; k++)
			est[i][j]+=b[k][j]*s[n][k][i]*area*thic;
		}
	}
}
	

/* function to form elsticty matrix D */
form_D(double v, double e, double d[][3])
{
double comm;
	comm=e/(1.-pow(v,2));
	d[0][0]=comm;
	d[0][1]=comm*v;
	d[0][2]=0.0;

	d[1][0]=comm*v;
	d[1][1]=comm;
	d[1][2]=0.;

	d[2][0]=0.;
	d[2][1]=0.;
	d[2][2]=comm*(1.0-v)/2.;
return;
}

/* function to read in the data from the primary data file */

rinput(char *file_sav)
{
char file[30];
FILE *fp;
int i,j,k,n;
double x1,y1,z1;
   
   printf("Enter the input file (.dat assumed)\n");
   gets(file); strcpy(file_sav, file); strcat(file,".dat");
   if((fp = fopen(file, "r")) == (FILE *) NULL)
   {printf("cannot open %s\n",file); exit(1);}
   fgets(buf,256,fp);
   fputs(buf,stdout);
   fgets(buf,256,fp);
   sscanf(buf,"%lf%lf",&dt,&damp);
   printf("%f %f \n",dt,damp);
   fgets(buf,256,fp);
   sscanf(buf,"%d %d %d %d %d %d %d %d %d %d %d",&nn,&ne,&nb,&nm,&nl,
   &ndime, &nstr, &nbe, &kmax,&nmax, &iln);

   fgets(buf,256,fp);
   sscanf(buf,"%d",&nq);


   if((cord = calloc(2*nn, sizeof(double))) == NULL)
   {printf(" out of core for cord\n");exit(1);}



   if((nod = calloc(4*ne, sizeof(int))) == NULL)
   {printf(" out of core for nod\n");exit(1);}


   for(i = 0; i < nn; i++)
   {
   fgets(buf,256,fp);
   sscanf(buf,"%d %lf %lf %lf",&n, &x1, &y1,&z1);
   cord[2*i] = x1;
   cord[2*i + 1] = y1;
   }



   if(ne > 0)
   {
      for(i = 0; i < ne; i++)
      {
      fgets(buf,256,fp);
      sscanf(buf,"%d %d %d %d %d",&n,&nod[4*i + 0],&nod[4*i + 1],
      &nod[4*i + 2],&nod[4*i + 3]);
      }
   }



   if(nl > 0)
   {
      for(i = 0; i < nl;i++)
      {
      fgets(buf,256,fp);
      }
   }



   if(nq > 0)
   {
      for(i = 0; i < nq; i++)
      {
      fgets(buf,256,fp);
      }
   }



   for(i = 0; i < nm; i++)
   {
   fgets(buf,256,fp);
   sscanf(buf,"%d %d %lf %lf %lf %lf %lf",&n,&mat[i],&amat[i][0],&amat[i][1],
   &amat[i][2],&amat[i][3],&amat[i][4]);
   }



   for(i = 0; i < nb; i++)
   {
   fgets(buf,256,fp);
   sscanf(buf,"%d %d %d %d",&nco[i][0],&nco[i][1],&nco[i][2],&nco[i][3]);
   }



   if(iln > 0)
   {
   iln = 0;
more:
   iln++;
   i = iln - 1;

   fgets(buf,256,fp);
   sscanf(buf,"%d %lf %lf %lf",&nc[i],&aload[i][0],&aload[i][1],&aload[i][2]);

   if(nc[i] < nn)
   goto more;
   }

   if(nstr > 0)
   {
   for(i = 0; i < nstr; i++)
   {
   fgets(buf,256,fp);
   fgets(buf,256,fp);
   }
   }
fclose(fp);
return;
}


/* function calculates the semi band width by finding the max diff
   between the node numbers */

   semi_band()
   {
   int k,l,n1,n2,ndiff;

   nsb = 0;
   for(k = 0; k < ne; k++)
   {
      n1 = nod[4*k + 0];
      for(l = 2; l >= 0; l--)
      {
      n2 = nod[4*k + l];
      ndiff = fabs(n1 - n2);
      if(ndiff > nsb) nsb = ndiff;
      n1 = n2;
      }
   }
   nsb = (nsb + 1)*2;
   return;
   }
