/* COPYRIGHTS ASAD IQBAL KHAN  */

/*                         RM001.C                                    */
/*      For Linking upto 10 Processors (ROOT+9) in 2 D array       */
#include <stdlib.h>
#include <chan.h>
#include <alt.h>
#include <thread.h>
#include <sema.h>
/*
|---------|--------------------|
| Channel | Descriptions       |
|---------|--------------------|
| 0-3     | BINDING PORTS      |
| in_p[0] | ID NUMBER          |
| in_p[1] | NUMBER OF COLUMNS  |
| 4-13    | MASTER TASK PORTS  |
| 14-23   | WORKER TASK PORTS  |
| 24-27   | PHYSICAL LINK PORTS|
|---------|--------------------|
*/
#define Cap             10240    /* Capacity of major buffers */
#define CapWorkSpool     2048    /* Capacity of worker spooling buffers */
#define LinkChanOffset  24
#define Nproc           10

SEMA lnk_sema0, lnk_sema1, lnk_sema2, lnk_sema3;
CHAN **in_p, **out_p, dummy, int_ch[10];

void  link_listener(int i);
void  link_spool_worker(int i);
void  link_spool_master_worker(int i);
void  master_listener(void);
void  worker_listener(void);
void  find_xy(int idno, int ncol, int *idx, int *idy);


main(argc, argv, envp, in_ports, ins, out_ports, outs)
int argc, ins, outs;
char *argv[], *envp[];
CHAN *in_ports[], *out_ports[];
{
register i;
int vali;

in_p = in_ports;
out_p = out_ports;

sema_init(&lnk_sema0,1);
sema_init(&lnk_sema1,1);
sema_init(&lnk_sema2,1);
sema_init(&lnk_sema3,1);


chan_init(&dummy);

for(i = 0; i < 10; i++)
chan_init(&int_ch[i]);

for(i = 0; i < 4; i++)
thread_create(link_listener,Cap+5120,1,i);

for(i = 1; i < 10; i++)     /*skip the 1st thread*/
thread_create(link_spool_worker,CapWorkSpool+3072,1,i);

thread_create(link_spool_master_worker,Cap+3072,1,0); /*1st thread*/

thread_create(master_listener,Cap+5120,0);
thread_create(worker_listener,Cap+5120,0);

alt_wait(1,&dummy);
return(1);
}

void  link_listener(int i)
{
char buf[Cap],bin;
int  msglen,sender,destination,vali,root;
int  ix, iy, ino, ncol;
int  idx, idy, idno;
int xdiff, ydiff;
div_t x;

i += LinkChanOffset;
idno =  (int ) in_p[0];
ncol = (int) in_p[1];

find_xy(idno,ncol,&idx,&idy);

for(;;)
{
alt_wait(1,in_p[i]);
chan_in_word(&msglen, in_p[i]);                 /* receive msglen in bytes */
/*buf = par_malloc(msglen);*/
chan_in_message(msglen,buf,in_p[i]);

ino = (int) buf[msglen-1];                     /* read destination no     */

find_xy(ino,ncol,&ix,&iy);

xdiff = ix-idx;     ydiff = iy-idy;

if(xdiff > 0)
{
sema_wait(&lnk_sema3);
chan_out_word(msglen, out_p[27]);     /*use link 3 to move towards Right*/
chan_out_message(msglen, buf, out_p[27]);
sema_signal(&lnk_sema3);
continue;                                /* alt_wait for next message      */

}

if(xdiff < 0)
{
sema_wait(&lnk_sema0);
chan_out_word(msglen, out_p[24]);     /*use link 0 to move towards Left */
chan_out_message(msglen, buf, out_p[24]);
sema_signal(&lnk_sema0);
continue;                                /* alt_wait for next message      */
}

if(ydiff > 0)
{
sema_wait(&lnk_sema2);
chan_out_word(msglen, out_p[26]);     /*use link 2 to move towards Down */
chan_out_message(msglen, buf, out_p[26]);
sema_signal(&lnk_sema2);
continue;                                /* alt_wait for next message      */
}

if(ydiff < 0)
{
sema_wait(&lnk_sema1);
chan_out_word(msglen, out_p[25]);     /*use link 1 to move towards Up   */
chan_out_message(msglen, buf, out_p[25]);
sema_signal(&lnk_sema1);
continue;                                /* alt_wait for next message      */
}

/* Well if none of the conditions have been met then this must be the
   Destination of the message.                                             */

sender = (int) buf[msglen-2];            /* Message from whom,  can have
                                            values 0...to...Nproc          */
root = (int) buf[msglen-3];              /* if root equal [1] then the
                                            Message is meant for the Master
                                            Okay                           */
if(root == 1)
{
msglen -= 3;
chan_out_word(msglen, out_p[sender+4]); /*use the approp. Master Channel  */
chan_out_message(msglen, buf, out_p[sender+4]);
continue;                                /* alt_wait for next message      */
}

/* This the destination of the message and its not for the Master Task
   then it must be for the Worker Task on this processor                   */
msglen -= 3;
chan_out_word(msglen, &int_ch[sender]); /*use the approp. Spool Channel */
chan_out_message(msglen, buf, &int_ch[sender]);
}                                        /* End of Forever Loop            */

}


void link_spool_worker(int i)
{
int msglen;
char buf[CapWorkSpool];

for(;;)
{
alt_wait(1,&int_ch[i]);
chan_in_word(&msglen, &int_ch[i]);
chan_in_message(msglen, buf, &int_ch[i]); /* spool */
chan_out_word(msglen, out_p[i+14]);
chan_out_message(msglen, buf, out_p[i+14]);
}
}

void link_spool_master_worker(int i)
{
int msglen;
char buf[Cap];

for(;;)
{
alt_wait(1,&int_ch[i]);
chan_in_word(&msglen, &int_ch[i]);
chan_in_message(msglen, buf, &int_ch[i]); /* spool */
chan_out_word(msglen, out_p[i+14]);
chan_out_message(msglen, buf, out_p[i+14]);
}
}


void  master_listener(void)
{
char buf[Cap],bin;
int  i, msglen_i, msglen,sender,destination,vali;
int  ix, iy, ino, ncol;
int  idx, idy, idno;
int xdiff, ydiff;
div_t x;

idno =  (int )in_p[0];
ncol = (int)  in_p[1];

find_xy(idno,ncol,&idx,&idy);

for(;;)
{
i = alt_wait(Nproc,in_p[4],in_p[5],in_p[6],in_p[7],in_p[8],in_p[9],
                         in_p[10],in_p[11],in_p[12],in_p[13]);
i += 4;
vali = (int) in_p[i];
chan_in_word(&msglen_i, in_p[i]);               /* receive msglen in bytes */
msglen = msglen_i+3;
/*buf = par_malloc(msglen);*/                    /* additional space for header*/
chan_in_message(msglen_i,buf,in_p[i]);

/* Form the Header Block fot the Message                                   */

ino = i-4;

buf[msglen-1] = (char) ino;    /*destination is determined from the channel
                                 used by the Master task                   */
buf[msglen-2] = (char) ino;   /*message from master masquerades as if processor
                              is sending itself the message therefore message
                              from master takes the idenity of the target*/

buf[msglen-3] = (char) 0;      /*make  root equal to [0] as message is not
                                 for the Master Task                       */

find_xy(ino,ncol,&ix,&iy);

xdiff = ix-idx;     ydiff = iy-idy;

if(xdiff > 0)
{
sema_wait(&lnk_sema3);
chan_out_word(msglen, out_p[27]);     /*use link 3 to move towards Right*/
chan_out_message(msglen, buf, out_p[27]);
sema_signal(&lnk_sema3);
continue;                                /* alt_wait for next message      */
}

if(xdiff < 0)
{
sema_wait(&lnk_sema0);
chan_out_word(msglen, out_p[24]);     /*use link 0 to move towards Left */
chan_out_message(msglen, buf, out_p[24]);
sema_signal(&lnk_sema0);
continue;                                /* alt_wait for next message      */
}

if(ydiff > 0)
{
sema_wait(&lnk_sema2);
chan_out_word(msglen, out_p[26]);     /*use link 2 to move towards Down */
chan_out_message(msglen, buf, out_p[26]);
sema_signal(&lnk_sema2);
continue;                                /* alt_wait for next message      */
}

if(ydiff < 0)
{
sema_wait(&lnk_sema1);
chan_out_word(msglen, out_p[25]);     /* use link 1 to move towards Up   */
chan_out_message(msglen, buf, out_p[25]);
sema_signal(&lnk_sema1);
continue;                              /* alt_wait for next message       */
}

/* Well if none of the conditions have been met then this must be the
   Destination of the message and as we know it has been sent by the
   Master Task so it is meant for the Worker Task.                        */

msglen -= 3;
chan_out_word(msglen, &int_ch[0]);  /*use the approp. Spool Worker Channel*/
chan_out_message(msglen, buf, &int_ch[0]);

/*par_free(buf);*/
}                                        /* End of Forever Loop            */

}


void  worker_listener(void)
{
char buf[Cap],bin;
register j;
int  i, msglen_i, msglen,sender,destination,vali;
int  ix, iy, ino, ncol;
int  idx, idy, idno;
int xdiff, ydiff;
div_t x;

idno =  (int )in_p[0];
ncol =  (int) in_p[1];

find_xy(idno,ncol,&idx,&idy);

for(;;)
{
i = alt_wait(Nproc,in_p[14],in_p[15],in_p[16],in_p[17],in_p[18],in_p[19],
                         in_p[20],in_p[21],in_p[22],in_p[23]);

i += 14;
vali = (int ) in_p[i];
chan_in_word(&msglen_i, in_p[i]);               /* receive msglen in bytes */
msglen = msglen_i+3;
/*buf = par_malloc(msglen);*/                   /* additional space for header*/
chan_in_message(msglen_i,buf,in_p[i]);

/* Form the Header Block fot the Message                                   */

ino = i-14;

buf[msglen-1] = (char) ino;    /*destination is determined from the channel
                                 used by the Worker task                   */
buf[msglen-2] = (char) idno;   /*copy the id no of this process            */

if((i-14) == idno)
buf[msglen-3] = (char) 1;      /*make root [1] due the port used by Worker */
else
buf[msglen-3] = (char) 0;      /*make root equal to [0] as message is not
                                 for the Master Task                       */

find_xy(ino,ncol,&ix,&iy);

xdiff = ix-idx;     ydiff = iy-idy;

if(xdiff > 0)
{
sema_wait(&lnk_sema3);
chan_out_word(msglen, out_p[27]);     /*use link 3 to move towards Right*/
chan_out_message(msglen, buf, out_p[27]);
sema_signal(&lnk_sema3);
continue;                                /* alt_wait for next message      */
}

if(xdiff < 0)
{
sema_wait(&lnk_sema0);
chan_out_word(msglen, out_p[24]);     /*use link 0 to move towards Left */
chan_out_message(msglen, buf, out_p[24]);
sema_signal(&lnk_sema0);
continue;                                /* alt_wait for next message      */
}

if(ydiff > 0)
{
sema_wait(&lnk_sema2);
chan_out_word(msglen, out_p[26]);     /*use link 2 to move towards Down */
chan_out_message(msglen, buf, out_p[26]);
sema_signal(&lnk_sema2);
continue;                                /* alt_wait for next message      */
}

if(ydiff < 0)
{
sema_wait(&lnk_sema1);
chan_out_word(msglen, out_p[25]);     /* use link 1 to move towards Up   */
chan_out_message(msglen, buf, out_p[25]);
sema_signal(&lnk_sema1);
continue;                              /* alt_wait for next message       */
}

/* Well if none of the conditions have been met then this must be the
   Destination of the message and as we know it has been sent by the
   Resident Worker Task so it is must be directed towards Master.
                                                                          */

msglen -= 3;
chan_out_word(msglen, out_p[4]);    /*use the approp. Master Channel */
chan_out_message(msglen, buf, out_p[4]);

/*par_free(buf);*/
}                                        /* End of Forever Loop            */

}

void find_xy(int idno, int ncol, int *idx, int *idy)
{
div_t x;

if(idno < ncol)
{
*idy = 0;  *idx = idno;
}
else
{
x = div(idno,ncol);   *idy = x.quot;     *idx = x.rem;
}
}

