
#include  <stdio.h>
#include  <math.h>
#include  <graphics.h>
#define   XX  extern             /*  set global variables to external */
#include  "plt.h"

float    *dfcord;

/* Produce Stress Plots */
dflnplot(float redfact)
{
    void   dflndat();
    void   pltdfkey(float);

    int    ie,in,node,is,i,n1,n2,n3,colour,corner,quad[8];
    float  dfmax,dffact,noddef,xcent,ycent,x[3],y[3];

    alloc_float(&dfcord, nn*ndime, "dfcord");

    dflndat();                     /* Read stress data from file         */

    dfmax = 0.0;                  /* Find max absolute stress           */
    for (in=0; in<nn; ++in)
    {
	noddef = dfcord[ndime*in+2] - cord[ndime*in+2];
	if (fabs(noddef) > fabs(dfmax))
	  dfmax = fabs(noddef);
    }
    dfmax *= 1.01;                 /* make sure stresses are within range */
    dffact = dfmax/5.5;
    pltdfkey(dffact);              /* Plot colour key to stress grades    */

    for (ie=0; ie<ne; ++ie) {      /* Find centre coordinates of element  */
      xcent = ycent = 0.0;
      for (i=0; i<3; ++i) {
	node = nod[4*ie+i] - 1;
	xcent += tcor[2*node+0];
	ycent += tcor[2*node+1];
      }
      xcent /= 3.0;
      ycent /= 3.0;
      for (is=0; is<3; ++is)
      {
	node = nod[4*ie+is] - 1;
	noddef = dfcord[ndime*node+2] - cord[ndime*node+2];
	if (noddef >= 0)
	  colour = (int)(noddef/dffact + 10.5);
	else
	  colour = (int)(noddef/dffact + 10.5);
	setcolor(colour);
	setfillstyle(SOLID_FILL,colour);
	n1 = nod[4*ie+is] - 1;
	switch (is)                        /* assign end nodes of sides */
	{
	  case 0:
	    n2 = nod[4*ie+1] - 1;
	    n3 = nod[4*ie+2] - 1;
	    break;
	  case 1:
	    n2 = nod[4*ie+2] - 1;
	    n3 = nod[4*ie+0] - 1;
	    break;
	  case 2:
	    n2 = nod[4*ie+0] - 1;
	    n3 = nod[4*ie+1] - 1;
	    break;
	  default:
	    printf("\n\n Error in DFLN.C  -- is = %d is not valid\n",is);
	    exit(1);
	}
	x[0] = (tcor[2*n1+0]+tcor[2*n2+0])/2.0;     /* midside points */
	y[0] = (tcor[2*n1+1]+tcor[2*n2+1])/2.0;
	x[1] = (tcor[2*n1+0]+tcor[2*n3+0])/2.0;
	y[1] = (tcor[2*n1+1]+tcor[2*n3+1])/2.0;
	x[2] =  tcor[2*n1+0];
	y[2] =  tcor[2*n1+1];
	if (redfact > 0.0) {
	  for (corner=0; corner<3; ++corner) {
	    x[corner] += (xcent-x[corner])*redfact;
	    y[corner] += (ycent-y[corner])*redfact;
	  }
	}
	quad[0] = (int)(cmxoff2 + sf*x[0]);
	quad[1] = (int)(twdowy - cmyoff2 - sf*y[0]);
	quad[2] = (int)(cmxoff2 + sf*xcent);
	quad[3] = (int)(twdowy - cmyoff2 - sf*ycent);
	quad[4] = (int)(cmxoff2 + sf*x[1]);
	quad[5] = (int)(twdowy - cmyoff2 - sf*y[1]);
	quad[6] = (int)(cmxoff2 + sf*x[2]);
	quad[7] = (int)(twdowy - cmyoff2 - sf*y[2]);

	fillpoly(4,quad);
      }
    }
return;
}



/*  read stress data from file  */
void dflndat(void)
{
		     /*  STRESS INPUT ROUTINE   */

    FILE    *fpdfin;
    int     i,n;
    char    line[180],dffile[20];

    strcpy(dffile,file);
    strcat(dffile,".dfl");
    fpdfin  = fopen(dffile,"r");


    /* ---------- Read Stresses --------- */
    if (nn > 0) {
      fprintf(fpout,"\n Deflected Coordinates : \n\n"
		   " Node               Coordinates \n");
      for (i = 0; i < nn; i++) {
	fgets  (line, 80, fpdfin);
	sscanf (line,"%5d",&n);
	sscanf (line,"%5d%f%f%f", &n,
		&dfcord[ndime*(n-1)+0], &dfcord[ndime*(n-1)+1], &dfcord[ndime*(n-1)+2]);
	fprintf(fpout,"%5d %25d %10d %10d\n",n,
		dfcord[ndime*(n-1)+0], dfcord[ndime*(n-1)+1], dfcord[ndime*(n-1)+2]);
      }
    }
    close(fpdfin);
}

/* Plot colour key to stress grades */
void  pltdfkey(float dffact)
{
    int   i,colour,grade,xpos,xinc,ypos,yinc;
    int   square[8];
    char  text[100],dummy[20];

    setcolor(1);  /* white */
    yinc = xinc = 10;
    xpos = 10 + xinc;
    for (grade=0; grade<11; ++grade)  {
      colour  =  grade + 5;
      setfillstyle(SOLID_FILL,colour);
      ypos = 25 + grade*yinc;
      square[0] = xpos;
      square[1] = ypos;
      square[2] = xpos+xinc;
      square[3] = ypos;
      square[4] = xpos+xinc;
      square[5] = ypos+yinc;
      square[6] = xpos;
      square[7] = ypos+yinc;
      fillpoly(4,square);

      if (grade<6)      /* Print Deflection Values */
	strcpy(text,gcvt((grade-5.5)*dffact,4,dummy));
      else
      {
	strcpy(text," ");
	strcat(text,gcvt((grade-5.5)*dffact,4,dummy));
      }
      text[8] = '\0';
      strcat(text,"  to  ");
      if (grade>=5)
	strcat(text," ");
      strcat(text,gcvt((grade-4.5)*dffact,4,dummy));
      text[22] = '\0';
      strcat(text," m");
      outtextxy(xpos+xinc+15,ypos,text);
    }
}