/* This program uses the main simple genetic code to optimize the partition
   of a FE mesh.  If choice is 0 then mesh is partitioned according to no of
   elems each side of mesh otherwise the area*me of elements.  A divider
   is set up and sga defines its optimum location. All global variables
   are for use in the objfunc(). The program reads INPUT.DAT and INPUT.ERR
   automatically. */

#define SHORT 1
#define BLUE_SKY 0   
#define Choice 1
#define HeuPmutation  0
#define Mag           5
#define SecOrder      0
#define Maxpop 150+1
#define Nvar     4+1
#define Maxchrom 16
#define Fmultiple 2
#define TOL       1.e-10
#define Pi        3.141592654
#define XX        extern

/* only difference from sga3.c is that it assigns unsigned type to all
chromosome type variable  this is an intermediate code for binary
crossover code  which will be implemented in sga5.c                        */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>

/***********objfunc***********/
XX int nn, ne, *nod, choice;
XX int ne1, *mark1, ne2, *mark2, *mark2_c;
XX double *cord;
XX double *hhi, *cg, *err, *hi;
XX double maxx,minx, maxy,miny,left_val,right_val,total_val;
XX int    div_hv;
XX double xseg1, yseg1, xseg2, yseg2;
XX int bsgt;
XX int *seg;
/*****************************/
typedef unsigned chromosome;

typedef struct{
              chromosome chrom[Nvar+1];
              double x[Nvar+1],objective,fitness;
              int parent1, parent2;
              int xsite;
              } ind_t;


typedef struct{
              int gen, jmax, jmin;
              chromosome chrom[Nvar+1];
              double x[Nvar+1],objective,fitness;
              int parent1, parent2;
              int xsite;
              } max_t;

/*****************************/

double objfunc(double *x);

void min_max(double *max_x,double *min_x,double *max_y,double *min_y);

void divider_1(double *x);

int  dirnode(double x1, double y1, double x2, double y2,
             double x3, double y3);

void divider_2(double x[]);

void calc_cg_elems(int i);

void elemcord_del(int n,int nnn[], double x[], double y[], double del[]);

void nodecord_del(int node, double *px, double *py, double *pdel);

void seglen(double x1,double y1, double x2, double y2, double *pL);

void interface(double  *L_int);

int match_mark2(int n1, int n2);

void calcang(double x[], double y[], double L[], double A[]);

void dot(double x1, double y1, double x2, double y2,
    double x3, double y3, double x4, double y4, double *pD);

int hoex15(double Yin[9], double Yout[1] );

int side21( double Yin[], double Yout[]);

void homogen_load(double *L, double *A, double *del, double *Yin);

void homogen_load_2(double *L, double *A, double *del, double *Yin);

void      initialize(int *,int *,int *,double *,double *,
                     ind_t *,double *,double *,double *,
                     double *,unsigned long *,unsigned long *,
                     int *, double *,max_t *max_ult,int);

void      initpop(int popsize,ind_t *oldpop,int lchrom,int nvar,
                  double *rvar);
double    chrom_map_x(unsigned chrom,double range1,double range2,int lchrom);

void      bit_on_off(unsigned *chrom, int i);

void      bit_on(unsigned *chrom, int i);

int       bit_check(unsigned chrom, int i);

int       flip(double probability);

double    random1(void);

double    objfunc(double *x);

void      initdata(int *popsize,int *lchrom,int *maxgen,double *pcross,
                   double *pmutation,unsigned long *nmutation,
                   unsigned long *ncross,int *nvar,
                   double *rvar);

void      initreport(int popsize,int lchrom,int maxgen,double pcross,
                     double pmutation,double max,double avg,double min,
                     double sumfitness);

void      generation(int popsize,ind_t *oldpop,
                     ind_t *newpop,int lchrom,unsigned long *ncross,unsigned long *nmutation,
                     double pcross,double pmutation,int nvar,double *rvar,
                     double avg, double sumfitness);

void      statistics(int popsize,double *max,double *avg,double *min,
                     double *sumfitness,ind_t*newpop,max_t *max_ult,
                     int nvar,int gen);
int       select(int popsize, double sumfitness, ind_t *old);

void      pre_select1(int popsize, double avg, ind_t *pop, int *choices);

int       select1(int *nremain,int *choices);

void      crossover(unsigned *parent1,unsigned *parent2,
                    unsigned *child1,unsigned *child2,
                    int lchrom,unsigned long *ncross,unsigned
                    long *nmutation,int *jcross,
                    double pcross,double pmutation,int nvar);

int       find_cross(int nvar, int lchrom, unsigned long *ncross,
                     double pcross);

void      variable_swap(unsigned parent1, unsigned parent2,
                        unsigned *child1, unsigned *child2);

void      bit_cross(unsigned parent1,unsigned parent2,int rt_off,
                    unsigned *child1,unsigned *child2);

int       round_(int j, int lchrom);

int       rnd(int low,int high);

int       mutation(double pmutation,unsigned long *nmutation);

void      report(int gen,ind_t *oldpop,ind_t *newpop,
                 int lchrom,double max,double avg,double min,double sumfitness,
                 unsigned long nmutation,unsigned long ncross, int popsize,
                 int nvar,max_t *max_ult);

void      scalepop(int popsize, double max, double avg, double min,
                   double *sumfitness,ind_t *pop);

void prescale(double umax, double uavg, double umin, double *a, double *b);

ind_t *oldpop, *newpop;

max_t max_ult;

int seq_step,max_step, Maxstep;
double max_prev;

int sga5(FILE *finf, int it)
{
char file[30];
int popsize, lchrom, maxgen, gen,nvar, jmax, jmin;
unsigned long ncross, nmutation;
double max,min,avg,pcross,pmutation,sumfitness,rvar[2*Nvar+1];
time_t t_start, t_finish, t_diff;

/* nvar-------------total number of design variables
   xvar[]-----------vector of the design variables
   rvar[]-----------interpolation ranges for design variables
   max_ult----------stores the maximum occurance of a raw fitness
*/



seq_step = 0;   /* initialize */
max_step = 0;   /* initialize */
max_prev = 0.;  /* initialize */

if(it == 0)
{                      
if((oldpop = calloc(Maxpop+1,sizeof(ind_t))) == NULL)
{printf("out of memory for oldpop[]\n"); exit(1);}

if((newpop = calloc(Maxpop+1,sizeof(ind_t))) == NULL)
{printf("out of memory for newpop[]\n"); exit(1);}
}


time(&t_start);

max_ult.gen = 1;

max_ult.objective = 0.;            /*initialize for maximum value*/

gen = 0;

initialize(&popsize,&lchrom,&maxgen,&pcross,&pmutation,oldpop,
           &max,&min,&avg,&sumfitness,&nmutation,&ncross,&nvar,
           rvar,&max_ult,gen);

do{
gen++; 

scalepop(popsize,max,avg,min,&sumfitness,oldpop);

generation(popsize,oldpop,newpop,lchrom,&ncross,&nmutation,
           pcross,pmutation,nvar,rvar,avg, sumfitness);

statistics(popsize,&max,&avg,&min,&sumfitness,newpop,&max_ult,nvar,gen);
#if SHORT
if(gen == 1 || gen == maxgen+1)
#endif
report(gen,oldpop,newpop,lchrom,max,avg,min,sumfitness,
       nmutation,ncross,popsize,nvar,&max_ult);
memcpy(oldpop, newpop, sizeof(newpop));

   jmin = max_ult.jmin;   /*hiearchial approach*/
   newpop[jmin].objective = max_ult.objective;
   newpop[jmin].fitness = max_ult.fitness;
   newpop[jmin].parent1 = max_ult.parent1; 
   newpop[jmin].parent2 = max_ult.parent2;
   newpop[jmin].xsite = max_ult.xsite;
   memcpy(&(newpop[jmin].chrom[0]),&(max_ult.chrom[0]),sizeof(chromosome)*(nvar+1));
   memcpy(&(newpop[jmin].x[0]),&(max_ult.x[0]),sizeof(double)*(nvar+1));

printf("Generation # %3d ult_fit %7.3f curr_max %7.3f prev_max %7.3f \n", gen, total_val,max, 
        max_ult.objective);
fprintf(finf,"Generation # %3d ult_fit %7.3f curr_max %7.3f prev_max %7.3f \n", gen, total_val,max, 
        max_ult.objective);


}
while((gen <= maxgen) && (max_step < Maxstep));

divider_1(max_ult.x); /*gives global mark1[] and mark2[]*/

time(&t_finish);   t_diff = t_finish - t_start;

printf("Total Time (mins.) = %f\n", (double)t_diff/60.);
printf("ne1 = %5d ....... ne2 = %5d\n",ne1, ne2);
printf("ult_fit = %12.3f  max_fit = %12.3f\n",total_val,max_ult.objective);
printf("The results are stored in ***genout.dat***\n");
fprintf(finf,"Total Time (mins.) = %f\n", (double)t_diff/60.);
fprintf(finf,"ne1 = %5d ....... ne2 = %5d\n",ne1, ne2);
fprintf(finf,"ult_fit = %12.3f  max_fit = %12.3f\n",total_val,max_ult.objective);
fprintf(finf,"The results are stored in ***genout.dat***\n");
return(1);
}


void initialize(int *popsize,int *lchrom,int *maxgen,double *pcross,
     double  *pmutation,ind_t *oldpop,double *max,
     double *min,double *avg,double *sumfitness,unsigned long *nmutation,
     unsigned long *ncross, int *nvar, double *rvar,max_t *max_ult,int gen)
{
initdata(popsize,lchrom,maxgen,pcross,pmutation,nmutation,ncross,
        nvar, rvar);

initpop(*popsize,oldpop,*lchrom,*nvar,rvar);

statistics(*popsize,max,avg,min,sumfitness,oldpop,max_ult,*nvar,gen);

initreport(*popsize,*lchrom,*maxgen,*pcross,*pmutation,
           *max,*avg,*min,*sumfitness);

}

void initdata(int *popsize,int *lchrom,int *maxgen,double *pcross,
              double *pmutation,unsigned long *nmutation,unsigned long *ncross,int *nvar,
              double *rvar)
{
register i,j;
int Lx, Ly;
FILE *fp;
char buf[120], dummy[50];
unsigned seed;

if((fp = fopen("sg5.var","r")) == (FILE *)NULL)
{printf("something wrong with sg5.var\n");exit(1);}


fgets(buf,120,fp); sscanf(buf,"%s %d",dummy,popsize);

fgets(buf,120,fp); sscanf(buf,"%s %d",dummy,lchrom);

fgets(buf,120,fp); sscanf(buf,"%s %d",dummy,maxgen);

fgets(buf,120,fp); sscanf(buf,"%s %lf",dummy,pcross);

fgets(buf,120,fp); sscanf(buf,"%s %lf",dummy,pmutation);

fgets(buf,120,fp); sscanf(buf,"%s %d",dummy,&seed);

fgets(buf,120,fp); sscanf(buf,"%s %d",dummy,nvar);

fgets(buf,120,fp); sscanf(buf,"%s %d",dummy,&Maxstep);


min_max(&maxx,&minx,&maxy,&miny); /* move limits */

rvar[1] = minx;  rvar[2] = maxx;
rvar[3] = miny;  rvar[4] = maxy;
rvar[5] = 0.;    rvar[6] = Pi;
rvar[7] = 0.;    rvar[8] = Pi;


srand(seed);

*nmutation = 0;

*ncross = 0;

fclose(fp);
}


void initpop(int popsize,ind_t *oldpop,int lchrom,int nvar,double *rvar)
{
register k,kk;
int j, j1, flip_t;

for(j = 1; j <= popsize; j++)                                /* zero lchrom*/
   for(k = 1; k <= nvar; oldpop[j].chrom[k++] = 0);

for(j = 1; j <= popsize; j++)
{
   for(k = 1; k <= nvar; k++)
      for(j1 = 1; j1 <= lchrom; j1++)
      {
      flip_t = flip(0.5);
      if(flip_t > 0) bit_on(&(oldpop[j].chrom[k]), j1);
      }


   for(k = 1; k <= nvar; k++)
   {
   kk = 2*k - 1;
   oldpop[j].x[k] = chrom_map_x(oldpop[j].chrom[k],
                               rvar[kk],rvar[kk+1],lchrom);
   }

   oldpop[j].objective = objfunc(oldpop[j].x);

   oldpop[j].parent1 = oldpop[j].parent2= oldpop[j].xsite = 0;

}
}

double chrom_map_x(unsigned chrom,double range1,double range2,int lchrom)
{
double diff,add_percent,res;

diff = range2-range1;
add_percent = ((double) chrom/ (pow(2.,lchrom)-1))*diff;
res = range1 + add_percent;

return(res);
}

int flip(double probability)
{
double random1_t;
if(probability == 1.0) return(1);

random1_t = random1();

if(random1_t <= probability) return(1);
else return(0);
}

double random1(void)
{
double result;
result = (double) rand()/RAND_MAX;
return(result);
}


void bit_on_off(unsigned *chrom, int i)
{
switch(i){
case 1:  *chrom ^= 01; break;  /*set on the 1st bit */
case 2:  *chrom ^= 02; break;
case 3:  *chrom ^= 04; break;
case 4:  *chrom ^= 010; break;
case 5:  *chrom ^= 020; break;
case 6:  *chrom ^= 040; break;
case 7:  *chrom ^= 0100; break;
case 8:  *chrom ^= 0200; break;
case 9:  *chrom ^= 0400; break;
case 10:  *chrom ^= 01000; break;
case 11: *chrom ^= 02000; break;
case 12: *chrom ^= 04000; break;
case 13: *chrom ^= 010000; break;
case 14: *chrom ^= 020000; break;
case 15: *chrom ^= 040000; break;
case 16: *chrom ^= 0100000; break;
case 17: *chrom ^= 0200000; break;
case 18: *chrom ^= 0400000; break;
case 19: *chrom ^= 01000000; break;
case 20: *chrom ^= 02000000; break;
case 21: *chrom ^= 04000000; break;
case 22: *chrom ^= 010000000; break;
case 23: *chrom ^= 020000000; break;
case 24: *chrom ^= 040000000; break;
case 25: *chrom ^= 0100000000; break;
case 26: *chrom ^= 0200000000; break;
case 27: *chrom ^= 0400000000; break;
case 28: *chrom ^= 01000000000; break;
case 29: *chrom ^= 02000000000; break;
case 30: *chrom ^= 04000000000; break;
case 31: *chrom ^= 010000000000; break;
}
}

void bit_on(unsigned *chrom, int i)
{
switch(i){
case 1:  *chrom |= 01; break;  /*set on the 1st bit */
case 2:  *chrom |= 02; break;
case 3:  *chrom |= 04; break;
case 4:  *chrom |= 010; break;
case 5:  *chrom |= 020; break;
case 6:  *chrom |= 040; break;
case 7:  *chrom |= 0100; break;
case 8:  *chrom |= 0200; break;
case 9:  *chrom |= 0400; break;
case 10: *chrom |= 01000; break;
case 11: *chrom |= 02000; break;
case 12: *chrom |= 04000; break;
case 13: *chrom |= 010000; break;
case 14: *chrom |= 020000; break;
case 15: *chrom |= 040000; break;
case 16: *chrom |= 0100000; break;
case 17: *chrom |= 0200000; break;
case 18: *chrom |= 0400000; break;
case 19: *chrom |= 01000000; break;
case 20: *chrom |= 02000000; break;
case 21: *chrom |= 04000000; break;
case 22: *chrom |= 010000000; break;
case 23: *chrom |= 020000000; break;
case 24: *chrom |= 040000000; break;
case 25: *chrom |= 0100000000; break;
case 26: *chrom |= 0200000000; break;
case 27: *chrom |= 0400000000; break;
case 28: *chrom |= 01000000000; break;
case 29: *chrom |= 02000000000; break;
case 30: *chrom |= 04000000000; break;
case 31: *chrom |= 010000000000; break;
}
}


int  bit_check(unsigned chrom, int i)
{
switch(i){
case 1:  if(chrom & 01)return(1); break;          /*if bit was on return [1]*/
case 2:  if(chrom & 02)return(1); break;
case 3:  if(chrom & 04)return(1); break;
case 4:  if(chrom & 010)return(1); break;
case 5:  if(chrom & 020)return(1); break;
case 6:  if(chrom & 040)return(1); break;
case 7:  if(chrom & 0100)return(1); break;
case 8:  if(chrom & 0200)return(1); break;
case 9:  if(chrom & 0400)return(1); break;
case 10:  if(chrom & 01000)return(1); break;
case 11: if(chrom & 02000)return(1); break;
case 12: if(chrom & 04000)return(1); break;
case 13: if(chrom & 010000)return(1); break;
case 14: if(chrom & 020000)return(1); break;
case 15: if(chrom & 040000)return(1); break;
case 16: if(chrom & 0100000)return(1); break;
case 17: if(chrom & 0200000)return(1); break;
case 18: if(chrom & 0400000)return(1); break;
case 19: if(chrom & 01000000)return(1); break;
case 20: if(chrom & 02000000)return(1); break;
case 21: if(chrom & 04000000)return(1); break;
case 22: if(chrom & 010000000)return(1); break;
case 23: if(chrom & 020000000)return(1); break;
case 24: if(chrom & 040000000)return(1); break;
case 25: if(chrom & 0100000000)return(1); break;
case 26: if(chrom & 0200000000)return(1); break;
case 27: if(chrom & 0400000000)return(1); break;
case 28: if(chrom & 01000000000)return(1); break;
case 29: if(chrom & 02000000000)return(1); break;
case 30: if(chrom & 04000000000)return(1); break;
case 31: if(chrom & 010000000000)return(1); break;
}
return(0);                               /* if bit is no on then return [0] */
}


void      statistics(int popsize,double *max,double *avg,double *min,
                 double *sumfitness,ind_t *pop,max_t *max_ult,int nvar,int gen)
{
int j,jmax, jmin;

jmax = 1;
jmin = 1;

*sumfitness = pop[1].objective;
*min        = pop[1].objective;
*max        = pop[1].objective;

for(j = 2; j <= popsize; j++)
{
*sumfitness += pop[j].objective;
if(pop[j].objective > *max)
  {
   *max = pop[j].objective;
   jmax = j;
  }

if(pop[j].objective < *min) 
  {
  *min = pop[j].objective;
  jmin = j;
  }

}

if(*max > max_ult->objective)
{
   max_step = 0; 

   max_ult->objective = *max;
   max_ult->gen = gen;
   max_ult->jmax = jmax;
   max_ult->jmin = jmin;
   max_ult->fitness = pop[jmax].fitness;
   max_ult->parent1 = pop[jmax].parent1;
   max_ult->parent2 = pop[jmax].parent2;
   max_ult->xsite = pop[jmax].xsite;
   memcpy(&(max_ult->chrom[0]),&(pop[jmax].chrom[0]), sizeof(chromosome)*(nvar+1));
   memcpy(&(max_ult->x[0]),&(pop[jmax].x[0]), sizeof(double)*(nvar+1));
}
else
{
max_step++;
printf("                     max_step = %3d\n",max_step);
}


*avg = *sumfitness/((double)popsize);
}

void      initreport(int popsize,int lchrom,int maxgen,double pcross,
                double pmutation,double max,double avg,double min,
                double sumfitness)
{
FILE *fpout;

if((fpout=fopen("genout.dat","w")) == (FILE *) NULL)
{printf("cannot open genout.dat\n");exit(1);}

fprintf(fpout,"Population size (popsize) = %5d\n",popsize);
fprintf(fpout,"Chromosome length (lchrom) = %5d\n",lchrom);
fprintf(fpout,"Maximum # of generations (maxgen) %5d\n",maxgen);
fprintf(fpout,"Crossover probability (pcross) = %10.5e\n",pcross);
fprintf(fpout,"Mutation probability (pmutation) = %10.5e\n",pmutation);
fprintf(fpout,"\n Initial Generation Statistics\n");
fprintf(fpout,"---------------------------------\n");
fprintf(fpout,"\n");
fprintf(fpout,"Initial population maximum fitness = %10.5e\n", max);
fprintf(fpout,"Initial population average fitness = %10.5e\n",avg);
fprintf(fpout,"Initial population minimum fitness = %10.5e\n",min);
fprintf(fpout,"Initial population sum of  fitness = %10.5e\n",sumfitness);
fprintf(fpout,"\n\n\n");
fclose(fpout);
}

void generation(int popsize,ind_t *oldpop,
     ind_t *newpop,int lchrom,unsigned long *ncross,unsigned long *nmutation,
     double pcross,double pmutation,int nvar, double *rvar,double avg, double sumfitness)
{
register k, kk;
int j, mate1, mate2, jcross,nremain, choices[2*Maxpop+1];

j = 1;

#if Choice
nremain = popsize;
printf("Enter pre_sel\n");
pre_select1(popsize, avg, oldpop, choices);
printf("Exit pre_sel\n");
#endif

do{

#if Choice
mate1 = select1(&nremain, choices);
mate2 = select1(&nremain, choices);
#else
mate1 = select(popsize, sumfitness, oldpop);
mate2 = select(popsize, sumfitness, oldpop);
#endif
crossover(oldpop[mate1].chrom, oldpop[mate2].chrom,
          &(newpop[j].chrom[0]), &(newpop[j+1].chrom[0]),
          lchrom,ncross,nmutation,&jcross,pcross,pmutation,nvar);

for(k = 1; k <= nvar; k++)
{
kk = 2*k - 1;
newpop[j].x[k] = chrom_map_x(newpop[j].chrom[k],rvar[kk],rvar[kk+1],lchrom);
}

newpop[j].objective = objfunc(newpop[j].x);
newpop[j].parent1 = mate1;
newpop[j].parent2 = mate2;

newpop[j].xsite = jcross;

for(k = 1; k <= nvar; k++)
{
kk = 2*k - 1;
newpop[j+1].x[k] = chrom_map_x(newpop[j+1].chrom[k],rvar[kk],rvar[kk+1],lchrom);
}
newpop[j+1].objective = objfunc(newpop[j+1].x);
newpop[j+1].parent1 = mate1;
newpop[j+1].parent2 = mate2;

newpop[j+1].xsite = jcross;

j += 2;
}
while(j < popsize);
}

int select(int popsize, double sumfitness, ind_t *pop)
{
int j;
double partsum, rand;

partsum = 0.; j = 0;
rand = random1()*sumfitness;
do{
j += 1;
partsum += pop[j].fitness;
}
while(partsum <= rand && j != popsize);
return(j);
}

void crossover(unsigned *parent1,unsigned *parent2,unsigned *child1,
               unsigned *child2,
               int lchrom,unsigned long *ncross,unsigned long *nmutation,int *jcross,
               double pcross,double pmutation, int nvar)
{
register i,j;
int test, quot,rem, rt_off;
div_t result;

#if HeuPmutation
pmutation *= ((double)((Maxstep+Mag*max_step)/Maxstep));
#endif

*jcross = find_cross(nvar, lchrom, ncross, pcross);

result = div(*jcross, lchrom); quot = result.quot;
rt_off=lchrom-result.rem+1;


memcpy(child1, parent1, sizeof(int)*(nvar+1));
memcpy(child2, parent2, sizeof(int)*(nvar+1));


for(i = 1; i <= nvar; i++)
{

   if(i == quot+1)
   bit_cross(parent1[i], parent2[i], rt_off, &child1[i], &child2[i]);
  
   if(i > quot+1)
   variable_swap(parent1[i], parent2[i], &child1[i], &child2[i]);

  for(j = 1; j <= lchrom; j++)
  {
  test = mutation(pmutation, nmutation);
           /*test = [0] no change , test = 1 jth bit is altered (mutated)*/
  if(test == 1)bit_on_off(&child1[i],j);

  test = mutation(pmutation, nmutation);
           /*test = [0] no change , test = 1 jth bit is altered (mutated)*/
  if(test == 1)bit_on_off(&child2[i],j);
  }
}

}



int find_cross(int nvar, int lchrom, unsigned long *ncross, double pcross)
{
int cross_over_point;
if(flip(pcross)  == 1)
{
cross_over_point = rnd(1, nvar*lchrom-1);
*ncross += 1;
}
else
cross_over_point = nvar*lchrom;


return(cross_over_point);
}

void variable_swap(unsigned parent1, unsigned parent2,
                   unsigned *child1, unsigned *child2)
{
*child1 = parent2;
*child2 = parent1;
}


void bit_cross(unsigned parent1,unsigned parent2,int rt_off,
               unsigned *child1,unsigned *child2)
{
unsigned a,b, siz;

siz = sizeof(unsigned)*8;

a = parent1; b = parent2;
a <<= rt_off;     a >>= rt_off;
b >>= siz-rt_off; b <<= siz-rt_off;
*child1 = a | b;

a = parent1; b = parent2;
a >>= siz-rt_off; a <<= siz-rt_off;
b <<= rt_off;     b >>= rt_off;
*child2 = a | b;

}


int round_(int j, int lchrom)
{

if(fmod(j,lchrom) == 0)
return(1);

return(0);
}

int  rnd(int low,int high)
{
int i;

if(low >= high) i = low;
else
{
i = (int) (random1()*(high-low+1) + low);
if(i > high) i = high;
}
return(i);
}

int  mutation(double pmutation,unsigned long *nmutation)
{
int mutate;

mutate = flip(pmutation);
if(mutate == 1)
{
*nmutation += 1;
return(1);
}
else
return(0);
}

void      report(int gen,ind_t *oldpop,ind_t *newpop,
          int lchrom,double max,double avg,double min,double sumfitness,
          unsigned long nmutation,unsigned long ncross, int popsize,
          int nvar,max_t *max_ult)
{
register k;
FILE *fpout;
int j;
double sum;

if((fpout=fopen("genout.dat","a")) == (FILE *) NULL)
{printf("cannot open genout.dat\n");exit(1);}

fprintf(fpout,"Generation #%5d data (lchrom %5d)\n",gen,lchrom);
fprintf(fpout,"No. parent1 parent2      x     fitness\n");
fprintf(fpout,"----------------------------------\n");

fprintf(fpout,"Old Generation of Generation No. %5d\n",gen);
for(j = 1; j <= popsize; j++)
{
fprintf(fpout,"%5d %5d %5d",j,oldpop[j].parent1,oldpop[j].parent2);
for(k=1; k <= nvar; k++)
    fprintf(fpout," %9.3e",oldpop[j].x[k]);
fprintf(fpout,"    %9.3e\n",oldpop[j].objective);
}

fprintf(fpout,"New Generation of Generation No. %5d\n",gen);
for(j = 1; j <= popsize; j++)
{
fprintf(fpout,"%5d %5d %5d",j,newpop[j].parent1,newpop[j].parent2);
for(k=1; k <= nvar; k++)
    fprintf(fpout," %9.3e",newpop[j].x[k]);
fprintf(fpout,"    %9.3e\n",newpop[j].objective);
}

fprintf(fpout,"New (%5d) generation's statistics\n",gen);

fprintf(fpout," max=%10.5e avg=%10.5e min=%10.5e sum=%10.5e\n nmutation=%d \
ncross=%d\n",max,avg,min,sumfitness,nmutation,ncross);

fprintf(fpout,"\nMax of all generations\n");
fprintf(fpout,"occurred in generation no.%5d \n",max_ult->gen);
fprintf(fpout,"chromosome variables\n");
for(k=1; k <= nvar; k++)
    fprintf(fpout," %5d",max_ult->chrom[k]);
fprintf(fpout,"\n");
fprintf(fpout,"design variables\n");
for(k=1; k <= nvar; k++)
    fprintf(fpout," %9.3e",max_ult->x[k]);
fprintf(fpout,"\n");
fprintf(fpout,"raw fitness = %f\n",max_ult->objective);
fclose(fpout);
}


void scalepop(int popsize, double max, double avg, double min,
              double *sumfitness,ind_t *pop)
{
register j;
double a, b;

prescale(max,avg,min,&a,&b);
*sumfitness = 0.;

for(j = 1; j <= popsize; j++)
{
pop[j].fitness = a*pop[j].objective+b;
if(pop[j].fitness < 0.) pop[j].fitness = 1.e-30; /*non-negetivity*/
*sumfitness += pop[j].fitness;
}
}

void prescale(double umax, double uavg, double umin, double *a, double *b)
{
double delta;

if(umin > (Fmultiple*uavg-umax)/(Fmultiple-1.)) /*Non-neg test*/
{
delta = umax-uavg;
if(delta == 0.) delta = .00000001;
*a = (Fmultiple-1.)*uavg/delta;
*b = uavg*(umax-Fmultiple*uavg)/delta;
}
else
{
delta = uavg - umin;
if(delta == 0.) delta = .00000001;
*a = uavg/delta;
*b = -umin*uavg/delta;
}
}

void pre_select1(int popsize, double avg, ind_t *pop, int *choices)
{
register  j, k, l;
int jassign, winner, fcount;
double expected, fraction[Maxpop+1], whole, maxfrac;

j = 0; k = 0;

do{
j++;
expected = pop[j].fitness/(avg+1.e-30);

fraction[j] = modf(expected, &whole);

jassign = (int) whole;

   while(jassign > 0)
   {
   k++; jassign--;
   choices[k] = j;
   }
}
while(j != popsize);

j = 0;
while(k < popsize)
{
j++;  

 maxfrac = 0.;
 fcount = 0;
 for(l = 1; l <= popsize; l++)
 {
 if(fraction[l] <= TOL) fcount++;
 maxfrac = (fraction[l] > maxfrac) ? fraction[l] : maxfrac;
 }

 if(fcount == popsize) /* prevent all fraction[] becoming too small or -ve */
 {
 printf("scaling up wrt to max fraction[j] to 0.5\n");
    for(l = 1; l <= popsize; l++)
    {
    if(maxfrac > 0.)
    fraction[l] = .5*fraction[l]/maxfrac; /* scale up to 0.5*/ 
    else
    fraction[l] = 1. - .5*fraction[l]/maxfrac;
    }
 }

 if(j > popsize) j = 1;

if(fraction[j] > 0.0)
  {
   winner = flip(fraction[j]);
   if(winner == 1)
     {
      k++;
      choices[k] = j;
      (fraction[j])--;
     }
   }
}
}


int select1(int *nremain,int *choices)
{
int jpick, index;

jpick = rnd(1, *nremain);

index = choices[jpick];

choices[jpick] = choices[*nremain];

*nremain = *nremain - 1;

return(index);
}




double objfunc(double *x)
{
register i;

static double fitness;
static long it;
int eno;
double  he, he_new, L_int, ratio;

if(it == 0)
{

cg = calloc(2*ne, sizeof(double));
if(cg == NULL){printf("out of core for cg\n");exit(1);}

mark1 = calloc(ne, sizeof(int));
if(mark1 == NULL){printf("out of core for mark1\n");exit(1);}

#if SecOrder
mark2_c = calloc(ne, sizeof(int));
if(mark2_c == NULL){printf("out of core for mark2_c[]\n");exit(1);}
#endif

mark2 = calloc(ne, sizeof(int));
if(mark2 == NULL){printf("out of core for mark2\n");exit(1);}

}
it++;

for(i=0; i < ne; i++)/*calc elem centre of gravities each time*/
calc_cg_elems(i);

divider_1(x);/*divide ne into ne1 & ne2 on left and right of divider*/

#if SecOrder
divider_2(x);
#endif

      left_val = 0.; right_val = 0.;

if(choice)
{

for(i = 0; i < ne1; i++)
{
eno = mark1[i]-1;
left_val += err[eno];/*the predicted no of elemnts*/
}


for(i = 0; i < ne2; i++)
{
eno = mark2[i]-1;
right_val += err[eno];
}
}
else
{
left_val = (double) ne1;
right_val = (double) ne2;
}

interface(&L_int);

total_val = left_val + right_val;

fitness = total_val - fabs(left_val-right_val)+TOL;

if(L_int < 1.e-4) L_int = 1.;

fitness = fitness - L_int;

if(fitness < 0.) fitness = TOL;

return(fitness);


}




/* function calculates the area of the triangular element */

void divider_1(double *x)
{
register i;
int test;
double Lmin, Lx, Ly, x1,y1,x2,y2;

/* allocate cg[2*ne], mark1[ne] and mark2[ne] arrays before hand.
   Elements left to divider are in mark1[] and on right in mark2[].        */
for(i = 0; i < ne; mark1[i++] = 0);
for(i = 0; i < ne; mark2[i++] = 0);

#if SecOrder
for(i = 0; i < ne; mark2_c[i++] = 0); /* for use in divider_2 */
#endif

Lx = maxx-minx; Ly = maxy-miny;

Lmin = (Lx > Ly) ? Ly : Lx;

x1 = x[1];  
y1 = x[2];
x2 = x1 + (Lmin/10.)*cos(x[3]);
y2 = y1 + (Lmin/10.)*sin(x[3]);

ne1 = ne2 = 0;
for(i = 0; i < ne; i++)
{

/*test=1 means that the element cg is left of divider_1(1,2)*/
test = dirnode(x1,y1,x2,y2,cg[2*i+0],cg[2*i+1]);
if(test == 1){ mark1[ne1] = i+1; (ne1)++;}
else
{ mark2[ne2] = i+1; (ne2)++;}

}

}


int dirnode(double x1, double y1, double x2, double y2,
            double x3, double y3)
{
double a11,a12,a21,a22,D,c1,c2,/*t,*/s;

/* Function dirnode returns 1 if the  node(x3,y3) is on the
left of the seg(1,2) else 0 */

   a11=x2-x1;
   a12=y2-y1;
   a21=y2-y1;
   a22 = -1.*(x2-x1);

   D=a11*a22-a12*a21;

   c1=x3-x1;
   c2=y3-y1;
   
   s=(a11*c2-a21*c1)/D;

   if(s < 0.)
   return(1);
   
   return(0);
}

void divider_2(double x[])
{
register i;
int eno, test, ne2_c;
double Lx, Ly, Lmin, x1,y1,x2,y2;

/*function checks that elements of mark2 are on the right of this new divider
as well if not they are removed from mark2 and added to mark1 */


Lx = maxx-minx; Ly = maxy-miny;

Lmin = (Lx > Ly) ? Ly : Lx;

x1 = x[1];  
y1 = x[2];
x2 = x1 + (Lmin/10.)*cos(x[4]);
y2 = y1 + (Lmin/10.)*sin(x[4]);


ne2_c = 0;
for(i = 0; i < ne2; i++)
{
eno = mark2[i]-1;
/*test=1 means that the element cg is left of divider*/
test = dirnode(x1,y1,x2,y2,cg[2*eno+0],cg[2*eno+1]);
if(test == 0)
  {
  mark2_c[ne2_c] = eno+1;  /* load this aux array */
  ne2_c++;
  }
  else
  { 
  mark1[ne1] = eno+1; /* add to mark1[] */
  ne1++;
  } 
}

for(i = 0; i < ne2_c; i++)
mark2[i] = mark2_c[i];

ne2 = ne2_c;

}

void calc_cg_elems(int i)
{
double sumx,sumy;
int nod1, nod2, nod3;


   nod1 = nod[4*i + 0];
   nod2 = nod[4*i + 1];
   nod3 = nod[4*i + 2];
   sumx=(cord[3*(nod1-1)+0]+cord[3*(nod2-1)+0]+cord[3*(nod3-1)+0]);
   sumy=(cord[3*(nod1-1)+1]+cord[3*(nod2-1)+1]+cord[3*(nod3-1)+1]);
   cg[2*i+0] = sumx/3.;   cg[2*i+1] = sumy/3.;

}

void min_max(double *max_x,double *min_x,double *max_y,double *min_y)
{
register i;
*min_x = 1.e30; *min_y = 1.e30; *max_x=0.; *max_y=0.;

for(i = 0; i < nn; i++)
{
*min_x=(*min_x < cord[3*i+0]) ? *min_x : cord[3*i+0];
*min_y=(*min_y < cord[3*i+1]) ? *min_y : cord[3*i+1];
*max_x=(*max_x > cord[3*i+0]) ? *max_x : cord[3*i+0];
*max_y=(*max_y > cord[3*i+1]) ? *max_y : cord[3*i+1];
}

}


void elemcord_del(int n,int nnn[], double x[], double y[], double del[])
{
register i;
   for(i = 0; i < 3; i++)
   {
   nnn[i] = nod[4*n+i];
   nodecord_del(nnn[i],&x[i], &y[i], &del[i]);
   }
}

void nodecord_del(int node, double *px, double *py, double *pdel)
{
   *px = cord[3*(node-1)+0];
   *py = cord[3*(node-1)+1];
   *pdel = hhi[node-1];
}


void seglen(double x1,double y1, double x2, double y2, double *pL)
{
*pL=sqrt(pow(x2-x1,2.)+pow(y2-y1,2.));
}


int interseg(double x1,double y1,double x2,double y2,
             double x3,double y3,double x4,double y4, double *x, double *y)
{
/* [-1] non-intersecting segments
   [0]  parallel segments
   [1]  intersecting segments     */

double a11=x2-x1;
double a12=x3-x4;
double a21=y2-y1;
double a22=y3-y4;
double c11=x3-x1;
double c21=y3-y1;
double D=a11*a22-a12*a21;
double t,s;



   if(fabs(D) < 1.e-10)
   return(0);                    /*lines are parallel*/

   t=(a22*c11-a12*c21)/D;
   s=(a11*c21-a21*c11)/D;

   *x=x1+t*(x2-x1);
   *y=y1+t*(y2-y1);

   if(fabs(t) <= 1.e-10) t = 1.e-10;

   if(fabs(s) <= 1.e-10) s = 1.e-10;

   if(fabs(t-1.) <= 1.e-10) t = .998;

   if(fabs(s-1.) <= 1.e-10) s = .998;

   if(t < .999 && t > 0. && s < .999 && s > 0.)
   return(1);                     /* lines intersect */
   else
   return(-1);
}



void interface(double  *L_int)
{
/* L_int is the number of cut edges at the interface*/
register i,j,k;
double edge, x1,y1,x2,y2, L, h_avg;
int ei,n1, n2, match;

*L_int = 0.;

for(i = 0; i < ne1; i++)
{
ei = mark1[i]-1;
   for(j = 0; j < 3; j++)
   {
   k = j+1; if(k > 2) k = 0;
   n1 = nod[4*ei+j];
   n2 = nod[4*ei+k];
   x1 = cord[3*(n1-1)+0]; y1 = cord[3*(n1-1)+1];
   x2 = cord[3*(n2-1)+0]; y2 = cord[3*(n2-1)+1];
   match = match_mark2(n1, n2);

   if(match == 1)
   {
   if(choice)
   {
   seglen(x1,y1,x2,y2,&L);
   h_avg = (hhi[n1-1] + hhi[n2-1])/2.;
   edge = L/h_avg;

   if(edge <= 0.) edge = 1.;
   *L_int += sqrt(edge); /* add an estimate on edges */    
   }
   else
   (*L_int)++;
   }
   }
}

}


int match_mark2(int n1, int n2)
{
register j, k, l;
int ej, nj1, nj2;

for(j = 0; j < ne2; j++)
{
ej = mark2[j]-1;

   for(k = 0; k < 3; k++)
   {
   l = k+1; if(l > 2) l = 0;
   nj1 = nod[4*ej+k];
   nj2 = nod[4*ej+l];
   if((nj1 == n1 && nj2 == n2) || (nj1 == n2 && nj2 == n1))
   return(1);                   /* match found*/
   }
}
return(0);
}



double neuron(double x[], double y[], double del[] )
{
double L[3],A[3], tot_A, Yin[9], Yout[1], rms;

     calcang(x,y,L,A);
     tot_A = A[0]+A[1]+A[2];

     if(fabs(tot_A - 180.) > 1.e-1)
     {
     fprintf(stdout,"*** Record Suspicious Sum_Angles = %f***\n", tot_A);
     exit(1);
     }
#if  BLUE_SKY
     homogen_load_2(L,A,del,Yin);
     side21(Yin, Yout);
#else
     homogen_load(L,A,del,Yin);
     hoex15(Yin, Yout);
#endif

return(Yout[0]);
}

void calcang(double x[], double y[], double L[], double A[])
{
int i,j,k;
double Aa,Bb,D;
   j = 0;
   k = 0;
   for(i=0; i < 3; i++)
   {

   j = i+1;   if(j > 2) j=0;

   k = j+1;   if(k > 2) k = 0;

   seglen(x[i],y[i],x[j],y[j],&Aa);
   seglen(x[j],y[j],x[k],y[k],&Bb);
   dot(x[j],y[j],x[i],y[i],x[j],y[j],x[k],y[k], &D);
   A[i]=acos(D/(Aa*Bb))*57.29577951;
   L[i] = Aa;
   }
}

void dot(double x1, double y1, double x2, double y2,
    double x3, double y3, double x4, double y4, double *pD)
{
   *pD = (x2-x1)*(x4-x3)+(y2-y1)*(y4-y3);
}


#if BLUE_SKY
void homogen_load_2(double *L, double *A, double *del, double *Yin)
{
double div;

div = del[0];
L[0]=L[0]/div;
L[1]=L[1]/div;
L[2]=L[2]/div;
del[0]=del[0]/div;
del[1]=del[1]/div;
del[2]=del[2]/div;
memmove(&Yin[0],&L[0], 3*sizeof(double));
memmove(&Yin[3],&A[0], 3*sizeof(double));
memmove(&Yin[6],&del[1], 2*sizeof(double));
}


/* Wed Mar 18 16:49:41 1992 (side21.c) Recall-Only Run-time for <side21> */
/*    Control Strategy is: <dbd> */
int side21( double Yin[], double Yout[])
{
    static double  Xout[21];  /* work arrays */
    long   ICmpT;  /* temp for comparisons */

    /* *** WARNING: Code generated assuming Recall = 0 *** */

    /* Read and scale input into network */
    Xout[2] = Yin[0] * (0.18411967) + (-1.0475949);
    Xout[3] = Yin[1] * (0.1303968) + (-1.0508156);
    Xout[4] = Yin[2] * (0.16693515) + (-1.0672749);
    Xout[5] = Yin[3] * (0.032501669) + (-2.0954135);
    Xout[6] = Yin[4] * (0.039786742) + (-2.20372);
    Xout[7] = Yin[5] * (0.027944278) + (-1.8454345);
    Xout[8] = Yin[6] * (0.27430327) + (-1.0311883);
    Xout[9] = Yin[7] * (0.34883314) + (-1.0618132);
LAB110:

    /* Generating code for PE 0 in layer 3 */
    Xout[10]  = (float)(0.54618168) + (float)(0.085533492) * Xout[2] +
        (float)(-0.21252041) * Xout[3] + (float)(-0.14860113) * Xout[4] +
        (float)(0.063799761) * Xout[5] + (float)(-0.17066513) * Xout[6] +
        (float)(0.046480015) * Xout[7] + (float)(0.55302745) * Xout[8] +
        (float)(0.41141197) * Xout[9];
    Xout[10] = tanh( Xout[10] );

    /* Generating code for PE 1 in layer 3 */
    Xout[11]  = (float)(-1.2205337) + (float)(0.50323206) * Xout[2] +
        (float)(0.88617492) * Xout[3] + (float)(0.68230879) * Xout[4] +
        (float)(-0.014477775) * Xout[5] + (float)(-0.11707368) * Xout[6] +
        (float)(0.12992224) * Xout[7] + (float)(-1.3911071) * Xout[8] +
        (float)(-1.1462131) * Xout[9];
    Xout[11] = tanh( Xout[11] );

    /* Generating code for PE 2 in layer 3 */
    Xout[12]  = (float)(-0.0086207502) + (float)(-0.050252583) * Xout[2] +
        (float)(0.11004429) * Xout[3] + (float)(0.15359177) * Xout[4] +
        (float)(0.029411305) * Xout[5] + (float)(0.10993723) * Xout[6] +
        (float)(-0.11219499) * Xout[7] + (float)(-0.14862484) * Xout[8] +
        (float)(0.039318699) * Xout[9];
    Xout[12] = tanh( Xout[12] );

    /* Generating code for PE 3 in layer 3 */
    Xout[13]  = (float)(4.0473428) + (float)(-0.25301269) * Xout[2] +
        (float)(-1.1034319) * Xout[3] + (float)(-0.2462479) * Xout[4] +
        (float)(-0.35364607) * Xout[5] + (float)(-0.43934938) * Xout[6] +
        (float)(-0.23387834) * Xout[7] + (float)(2.6548531) * Xout[8] +
        (float)(2.1599898) * Xout[9];
    Xout[13] = tanh( Xout[13] );

    /* Generating code for PE 4 in layer 3 */
    Xout[14]  = (float)(0.19829604) + (float)(0.09846554) * Xout[2] +
        (float)(0.098290749) * Xout[3] + (float)(0.10703668) * Xout[4] +
        (float)(0.04930957) * Xout[5] + (float)(0.13335407) * Xout[6] +
        (float)(-0.2694723) * Xout[7] + (float)(0.07709188) * Xout[8] +
        (float)(0.06700629) * Xout[9];
    Xout[14] = tanh( Xout[14] );

    /* Generating code for PE 5 in layer 3 */
    Xout[15]  = (float)(0.044349976) + (float)(-0.045579452) * Xout[2] +
        (float)(0.0021004765) * Xout[3] + (float)(-0.023146909) * Xout[4] +
        (float)(0.04160523) * Xout[5] + (float)(0.00013204906) * Xout[6] +
        (float)(0.081974357) * Xout[7] + (float)(0.0069482052) * Xout[8] +
        (float)(0.072045647) * Xout[9];
    Xout[15] = tanh( Xout[15] );

    /* Generating code for PE 6 in layer 3 */
    Xout[16]  = (float)(0.1370659) + (float)(0.088330001) * Xout[2] +
        (float)(0.087633923) * Xout[3] + (float)(0.01764252) * Xout[4] +
        (float)(0.042855486) * Xout[5] + (float)(0.088984802) * Xout[6] +
        (float)(-0.15256917) * Xout[7] + (float)(0.18512784) * Xout[8] +
        (float)(-0.072471261) * Xout[9];
    Xout[16] = tanh( Xout[16] );

    /* Generating code for PE 7 in layer 3 */
    Xout[17]  = (float)(-0.21536033) + (float)(-0.0076851728) * Xout[2] +
        (float)(-0.041946534) * Xout[3] + (float)(-0.1593183) * Xout[4] +
        (float)(-0.082473248) * Xout[5] + (float)(-0.11573834) * Xout[6] +
        (float)(0.13915163) * Xout[7] + (float)(-0.12061974) * Xout[8] +
        (float)(-0.10466938) * Xout[9];
    Xout[17] = tanh( Xout[17] );

    /* Generating code for PE 8 in layer 3 */
    Xout[18]  = (float)(-0.60948455) + (float)(-0.084934995) * Xout[2] +
        (float)(-0.039615232) * Xout[3] + (float)(-0.048249222) * Xout[4] +
        (float)(-0.08348643) * Xout[5] + (float)(-0.034156118) * Xout[6] +
        (float)(0.33190176) * Xout[7] + (float)(-0.5521968) * Xout[8] +
        (float)(-0.32536831) * Xout[9];
    Xout[18] = tanh( Xout[18] );

    /* Generating code for PE 9 in layer 3 */
    Xout[19]  = (float)(-1.085698) + (float)(-0.61250514) * Xout[2] +
        (float)(-0.48786807) * Xout[3] + (float)(-0.43282017) * Xout[4] +
        (float)(0.27961707) * Xout[5] + (float)(0.37657499) * Xout[6] +
        (float)(-0.21017857) * Xout[7] + (float)(0.13068153) * Xout[8] +
        (float)(0.015385672) * Xout[9];
    Xout[19] = tanh( Xout[19] );

    /* Generating code for PE 0 in layer 4 */
    Xout[20]  = (float)(0.15488768) + (float)(0.0087332204) * Xout[10] +
        (float)(0.27194756) * Xout[11] + (float)(0.026975052) * Xout[12] +
        (float)(-0.92414081) * Xout[13] + (float)(0.073765166) * Xout[14] +
        (float)(-0.011192123) * Xout[15] + (float)(0.075974852) * Xout[16] +
        (float)(-0.059688285) * Xout[17] + (float)(-0.10668954) * Xout[18] +
        (float)(-0.27624416) * Xout[19];
    Xout[20] = tanh( Xout[20] );

    /* De-scale and write output from network */
    Yout[0] = Xout[20] * (63.749999) + (52);
    return( 0 );
}
#else

/* Wed Feb 19 21:00:13 1992 (hoex15.c) Recall-Only Run-time for <heox15> */
/*    Control Strategy is: <dbd> */
int hoex15(double Yin[9], double Yout[1] )
{
    static double  Xout[17];  /* work arrays */
    long   ICmpT;  /* temp for comparisons */

    /* *** WARNING: Code generated assuming Recall = 0 *** */

    /* Read and scale input into network */
    Xout[2] = Yin[0] * (0.18411967) + (-1.0475949);
    Xout[3] = Yin[1] * (0.13056705) + (-1.0534933);
    Xout[4] = Yin[2] * (0.16848206) + (-1.0864313);
    Xout[5] = Yin[3] * (0.032501669) + (-2.0954135);
    Xout[6] = Yin[4] * (0.043108462) + (-2.3042163);
    Xout[7] = Yin[5] * (0.027944278) + (-1.8454345);
    Xout[8] = Yin[6] + (-2);
    Xout[9] = Yin[7] * (0.27430327) + (-1.0311883);
    Xout[10] = Yin[8] * (0.34883314) + (-1.0618132);
LAB110:

    /* Generating code for PE 0 in layer 3 */
    Xout[11]  = 0.15789863 + 0.10383216 * Xout[2] + (-0.22523752) * Xout[3] +
        (-0.67347956) * Xout[4] + (-0.05143927) * Xout[5] +
        (-0.3063916) * Xout[6] + 0.14410186 * Xout[7] +
        (-0.053388517) * Xout[8] + 0.49070397 * Xout[9] + 0.05235222 * Xout[10]
        ;
    Xout[11] = tanh( Xout[11] );

    /* Generating code for PE 1 in layer 3 */
    Xout[12]  = (-7.2727799) + 1.0029693 * Xout[2] + 1.1253526 * Xout[3] +
        (-0.096689932) * Xout[4] + 0.15600094 * Xout[5] + 0.10614354 * Xout[6]
         + 0.061088312 * Xout[7] + (-0.035957791) * Xout[8] +
        (-6.1995211) * Xout[9] + (-2.2345977) * Xout[10];
    Xout[12] = tanh( Xout[12] );

    /* Generating code for PE 2 in layer 3 */
    Xout[13]  = (-1.8205116) + (-0.38261673) * Xout[2] + 1.2245123 * Xout[3] +
        1.0399791 * Xout[4] + 0.10367764 * Xout[5] + 0.22796212 * Xout[6] +
        0.14192601 * Xout[7] + (-0.079842076) * Xout[8] +
        (-1.2307496) * Xout[9] + (-1.8718977) * Xout[10];
    Xout[13] = tanh( Xout[13] );

    /* Generating code for PE 3 in layer 3 */
    Xout[14]  = (-0.93488657) + (-0.95023084) * Xout[2] +
        (-0.054171421) * Xout[3] + (-0.3570905) * Xout[4] +
        0.046458505 * Xout[5] + 0.1223551 * Xout[6] + (-0.21982495) * Xout[7]
         + 0.065754555 * Xout[8] + (-0.10536617) * Xout[9] +
        (-0.051805291) * Xout[10];
    Xout[14] = tanh( Xout[14] );

    /* Generating code for PE 4 in layer 3 */
    Xout[15]  = 0.010340533 + (-0.13679253) * Xout[2] + 0.19975138 * Xout[3] +
        0.16966611 * Xout[4] + 0.099696361 * Xout[5] + 0.13113278 * Xout[6] +
        0.02222888 * Xout[7] + 0.061033741 * Xout[8] + (-0.047290441) * Xout[9]
         + (-0.058939323) * Xout[10];
    Xout[15] = tanh( Xout[15] );

    /* Generating code for PE 0 in layer 4 */
    Xout[16]  = 0.11256779 + (-0.13167156) * Xout[11] + 0.77680171 * Xout[12]
         + 0.36559567 * Xout[13] + (-0.37231231) * Xout[14] +
        0.033708036 * Xout[15];
    Xout[16] = tanh( Xout[16] );

    /* De-scale and write output from network */
    Yout[0] = Xout[16] * (63.749999) + (52);
    return( 0 );
}


void homogen_load(double *L, double *A, double *del, double *Yin)
{
double div;

div = del[0];
L[0]=L[0]/div;
L[1]=L[1]/div;
L[2]=L[2]/div;
del[0]=del[0]/div;
del[1]=del[1]/div;
del[2]=del[2]/div;
memmove(&Yin[0],&L[0], 3*sizeof(double));
memmove(&Yin[3],&A[0], 3*sizeof(double));
memmove(&Yin[6],&del[0], 3*sizeof(double));
}

#endif

